@@ -1,8 +1,12 @@
-from starlette.requests import Request
-from fastapi import FastAPI
-from apish.metadata import Metadata
-from apish.routes import RouteGenerator, Namespace, Resource
+from fastapi import FastAPI, Request
+from fastapi.encoders import jsonable_encoder
+from fastapi.responses import JSONResponse
+from fastapi.exception_handlers import RequestValidationError
+from starlette.exceptions import HTTPException
 
+from apish.base import Problem
+from apish.metadata import Metadata
+from apish.routes import Namespace, Resource, RouteGenerator
 
 ns = Namespace(["Generic"])
 
@@ -21,6 +25,18 @@ class Health(Resource):
         return {"status": "ok"}
 
 
+async def from_validation_error(_request: Request, exc: RequestValidationError):
+    body = jsonable_encoder(Problem(title="Validation error", status=400, detail=exc.errors()))
+    return JSONResponse(status_code=400, content=body)
+
+
+async def from_http_exception(_request: Request, exc: HTTPException):
+    problem = Problem(title=exc.detail, status=exc.status_code, detail={})
+    body = jsonable_encoder(problem)
+    # pylint: disable=no-member
+    return JSONResponse(status_code=problem.status, content=body)
+
+
 class Application(FastAPI):
     def __init__(self, root: str, metadata: Metadata, **kwargs):
         self.root = root
@@ -34,6 +50,8 @@ class Application(FastAPI):
             redoc_url=None,
             **kwargs,
         )
+        self.add_exception_handler(RequestValidationError, from_validation_error)
+        self.add_exception_handler(HTTPException, from_http_exception)
         self.add(ns)
 
     def add(self, route_generator: RouteGenerator) -> None:
